import math

from geo import c


def trop_modifisert_hopfield(
        elevasjon_rad: float
) -> float:
    """
    Implementasjon av modifisert Hopfield modell for beregning av troposfæriske korreksjoner.

    :param elevasjon_rad: float. elevasjons vinkel i radians
    :return: float. korreksjonsverdi i meter.
    """
    T0 = 291.15
    e0 = 13.82
    P0 = 1013.25
    hw = 11000
    hd = 40136 + 148.72 * (T0 - 273.16)
    c1 = 77.64
    c2 = -12.96
    c3 = 371800
    Re = 6371000
    Nd0 = c1 * P0 / T0
    Nw0 = c2 * e0 / T0 + c3 * e0 / T0 ** 2

    ad = -math.sin(elevasjon_rad) / hd
    bd = -math.cos(elevasjon_rad) ** 2 / (2 * hd * Re)

    alfa_1d = 1
    alfa_2d = 4 * ad
    alfa_3d = 6 * ad ** 2 + 4 * bd
    alfa_4d = 4 * ad * (ad ** 2 + 3 * bd)
    alfa_5d = ad ** 4 + 12 * ad ** 2 * bd + 6 * bd ** 2
    alfa_6d = 4 * ad * bd * (ad ** 2 + 3 * bd)
    alfa_7d = bd ** 2 * (6 * ad ** 2 + 4 * bd)
    alfa_8d = 4 * ad * bd ** 3
    alfa_9d = bd ** 4

    alfa_d = [alfa_1d, alfa_2d, alfa_3d, alfa_4d, alfa_5d, alfa_6d, alfa_7d, alfa_8d, alfa_9d]

    aw = -math.sin(elevasjon_rad) / hw
    bw = -math.cos(elevasjon_rad) ** 2 / (2 * hw * Re)

    alfa_1w = 1
    alfa_2w = 4 * aw
    alfa_3w = 6 * aw ** 2 + 4 * bw
    alfa_4w = 4 * aw * (aw ** 2 + 3 * bw)
    alfa_5w = aw ** 4 + 12 * aw ** 2 * bw + 6 * bw ** 2
    alfa_6w = 4 * aw * bw * (aw ** 2 + 3 * bw)
    alfa_7w = bw ** 2 * (6 * aw ** 2 + 4 * bw)
    alfa_8w = 4 * aw * bw ** 3
    alfa_9w = bw ** 4

    alfa_w = [alfa_1w, alfa_2w, alfa_3w, alfa_4w, alfa_5w, alfa_6w, alfa_7w, alfa_8w, alfa_9w]

    rd = math.sqrt((Re + hd) ** 2 - (Re * math.cos(elevasjon_rad)) ** 2) - Re * math.sin(elevasjon_rad)
    rw = math.sqrt((Re + hw) ** 2 - (Re * math.cos(elevasjon_rad)) ** 2) - Re * math.sin(elevasjon_rad)

    delta_trop_d = 1e-6 * Nd0 * (math.fsum([rd ** k * alfa / k for alfa, k in zip(alfa_d, range(1, 10))]))
    delta_trop_w = 1e-6 * Nw0 * (math.fsum([rw ** k * alfa / k for alfa, k in zip(alfa_w, range(1, 10))]))

    return delta_trop_d + delta_trop_w  # TODO vurder bruk av Niell MF


def ion_klobuchar(
        lat, lon,
        asimut_rad, elevasjon_rad,
        current_time_gps_sec,
        alpha, beta
):
    """
    Implementasjon av Klobuchar modell for beregning av ionosfæriske korreksjoner.

    :param lat: float. Latitude i radians
    :param lon: float. Longitude i radians
    :param asimut_rad: float. asimut i radians
    :param elevasjon_rad: float. elevasjonsvinkel i radias
    :param current_time_gps_sec: int. Antall GPS sekunder.
    :param alpha: list. Alpha koder fra Navigasjonsfil
    :param beta: list. Beta koder fra Navigasjonsfil
    :return: float. korreksjonsverdi i meter
    """
    rad2semi = 1 / math.pi
    semi2rad = math.pi

    # 1. beregne jord-sentrisk vinkel i halvsirkler
    psi = 0.0137 / (elevasjon_rad * rad2semi + 0.11) - 0.022  # halvsirkler

    # 2. beregne lat for Ionospheric Pierce Point
    lat_IPP = lat * rad2semi + psi * math.cos(asimut_rad)  # halvsirkler
    if lat_IPP > 0.416:
        lat_IPP = 0.416
    elif lat_IPP < -0.416:
        lat_IPP = -0.416

    # 3. bergne lon for Ionospheric Pierce Point
    lon_IPP = lon * rad2semi + psi * math.sin(asimut_rad) / math.cos(lat_IPP * semi2rad)

    # 4. geomagnetisk lat IPP
    geo_lat_IPP = lat_IPP + 0.064 * math.cos((lon_IPP - 1.617) * semi2rad)

    # 5. IPP lokal tid
    t = 43200 * lon_IPP + current_time_gps_sec
    if t >= 86400:
        t -= 86400
    elif t < 0:
        t += 86400
    # 6. Amplitude of ionospheric delay
    Amp = math.fsum([(geo_lat_IPP ** n) * alpha_i for alpha_i, n in zip(alpha, range(4))])
    if Amp < 0:
        Amp = 0

    # 7. Period of ionospheric delay
    Per = math.fsum([(geo_lat_IPP ** n) * beta_i for beta_i, n in zip(beta, range(4))])
    if Per < 72000:
        Per = 72000

    # 8. Phase of ionospheric delay
    X_I = 2 * math.pi * (t - 50400) / Per

    # 9. Slant factor
    F = 1 + 16 * (0.53 - elevasjon_rad * rad2semi) ** 3

    # 10. Ionospheric time delay (seconds)
    if math.fabs(X_I) <= 1.57:
        Ion = F * (5e-9 + Amp * (1 - 0.5 * X_I ** 2 + (X_I ** 4) / 24))
    else:
        Ion = 5e-9 * F

    return Ion * c  # meters
