import os
from datetime import timedelta
from typing import Dict

from gps_time import GPSTime
from nmbu.rinex.reader import read_rinex_file
from pyproj import Transformer
from tabulate import tabulate

from corrections import trop_modifisert_hopfield, ion_klobuchar
from geo import *
from spp_utils import *


def spp(
        obs_file: str,
        nav_file: str,
        gnss: Dict[str, str],
        start_tidspunkt: datetime,
        slutt_tidspunkt: datetime,
        fil_for_mellomresultater: str,
        fil_for_hoved_resultater: str,
        trop_korreksjoner: bool = True,
        ion_korreksjoner: bool = True,
) -> (float, float, float):
    """
    Denne funksjonen beregner koordinater til mottaker, gitt navigasjons- og observasjons filer i RINEX 3 eller 4 format.
    Følgende biblioteker må være tilgjengelig:
    - nmbu.rinex.reader
    - numpy
    - pyproj
    - tabulate
    - gps_time

    Metoden vil bruke spesifiserte GNSS og kodeobservasjoner for beregninger.
    Brukeren har også mulighet til å slå av/på troposfæriske og/eller ionosfæriske korreksjoner.
    For troposfæriske korreksjoner benyttes modifisert Hopfield modell.
    For ionosfæriske korreksjoner benyttes Klobuchar modell.

    Funksjonen vil returnere (x,y,z) til mottaker og samtidig produsere Resultater.txt og MellomResultater.txt,
    som vil inneholde resultater av diverse beregninger underveis i iterasjonene.

    :param obs_file: str. Sti til RINEX observasjonsfil.
    :param nav_file: str. Sti til RINEX navigasjonsfil.
    :param gnss: Dict. Liste med GNSS og tilhørende kodeobservasjoner. For eksempel gnss={'G': 'C1C', 'E': 'C1X'},
    :param start_tidspunkt: datetime. Start for tidsinterval for beregninger.
    :param slutt_tidspunkt: datetime. Slutt for tidsinterval for beregninger.
    :param fil_for_mellomresultater: str. Sti til fil for mellomresultater.
    :param fil_for_hoved_resultater: str. Sti til fil for hoved resultater.
    :param trop_korreksjoner: bool. Flag for å bruke troposfæriske korreksjoner.
    :param ion_korreksjoner: bool. Flag for å bruke ionosfæriske korreksjoner.
    :return: (float, float, float). X, Y, Z ECEF koordinater til mottaker.
    """
    with open(fil_for_mellomresultater, 'w') as mellomresultater, open(fil_for_hoved_resultater, 'w') as resultater:
        mellomresultater.write("Absolutt-posisjonsbestemmelse med GPS C1C kodeobservasjoner\n")
        mellomresultater.write("\n")
        mixed_gnss = len(gnss) > 1

        resultater.write("Absolutt-posisjonsbestemmelse med GPS C1C kodeobservasjoner\n")
        resultater.write("\n")
        data_to_resultater = []

        nav_data = read_rinex_file(nav_file)
        obs_data = read_rinex_file(obs_file,
                                   gnss=list(gnss.keys()),
                                   obs_types=list(gnss.values()),
                                   start_epoch=start_tidspunkt.isoformat(),
                                   end_epoch=slutt_tidspunkt.isoformat()
                                   )

        if len(obs_data.data.satellites) == 0:
            raise ValueError("RINEX Observation data seem to be missing. "
                             "Make sure that time interval and other applied filters match provided RINEX files.")

        interval = timedelta(seconds=obs_data.header.interval)

        current_time = start_tidspunkt

        while current_time <= slutt_tidspunkt:
            # foreløpige verdier for ukjente parametre
            x_st0 = obs_data.header.approximate_position['X']
            y_st0 = obs_data.header.approximate_position['Y']
            z_st0 = obs_data.header.approximate_position['Z']
            lon, lat, alt = Transformer.from_proj(ecef, lla).transform(x_st0, y_st0, z_st0, radians=True)
            dt0 = 0  # antatt verdi, skal justeres etterhvert
            dt_system = 0

            current_gps_time = GPSTime.from_datetime(current_time)
            all_svs_for_time = {sv: bl for sv, bl in obs_data.data.satellites.items() if
                                current_time.isoformat() in bl.keys()}

            sv_list = list(all_svs_for_time.keys())
            R_ij = [
                beregne_R_ji(
                    nav_block=nav_data.find_closest_match(sv=sv, timestamp=current_time.isoformat()),
                    obs_block=all_svs_for_time,
                    sv=sv,
                    current_gps_time=current_gps_time,
                    current_time=current_time
                )
                for sv in all_svs_for_time.keys()
            ]
            # check for empty values in R_ij and remove corresponding SVs
            for i in range(len(R_ij)):
                if math.isnan(R_ij[i]):
                    del sv_list[i]
            R_ij = [r for r in R_ij if not math.isnan(r)]
            T_j = [current_gps_time - R_value / c for R_value in R_ij]

            forbedring = 10
            while forbedring > 0.1:
                R_ij_korr = []
                sat_koord = []
                Trel_m = []
                maaletid = []
                asimut = []
                elevasjon = []
                removed_indexes = []

                for sv, tj, index in zip(sv_list, T_j, range(len(T_j))):
                    x_sv, y_sv, z_sv, tk, Trel_i = beregne_sv_koordinater(
                        block=nav_data.find_closest_match(sv=sv, timestamp=current_time.isoformat()),
                        tidspunkt=tj,
                        mot_koord=(x_st0, y_st0, z_st0),
                        gnss = sv_to_gnss(sv)
                    )
                    asim, elev = asimut_og_elevasjon((x_st0, y_st0, z_st0), (x_sv, y_sv, z_sv), sv_to_gnss(sv))
                    trop = trop_modifisert_hopfield(elev) if trop_korreksjoner else 0
                    if ion_korreksjoner:
                        # Vi antar at atmosfærenspåvirkning er omtrent lik på alle satellittsignaler.
                        # Med denne antagelse kan vi bruke Klobuchar til Galileo satellittene.
                        # Siden navigasjonsmelding til Galileo inneholder ionosfæriske korreksjoner
                        # til NEQUICK modell, kan vi ikke bruke de til Klobuchar.
                        # Dermed skal korreksjonene fra GPS navmelding brukes.
                        closest_correction_block = nav_data.find_closest_correction_match(
                            correction_type="ION",
                            sv='G', timestamp=current_time.isoformat()
                        )
                        if not closest_correction_block:
                            ion = 0
                        else:
                            ion = ion_klobuchar(
                                lat, lon,
                                asim, elev,
                                current_gps_time.seconds,
                                alpha=[closest_correction_block.Alpha0, closest_correction_block.Alpha1,
                                       closest_correction_block.Alpha2, closest_correction_block.Alpha3],
                                beta=[closest_correction_block.Beta0, closest_correction_block.Beta1,
                                      closest_correction_block.Beta2, closest_correction_block.Beta3]
                            )
                    else:
                        ion = 0
                    if elev >= math.radians(15):
                        sat_koord.append((x_sv, y_sv, z_sv))
                        Trel_m.append(Trel_i)
                        maaletid.append(tk)
                        asimut.append(asim)
                        elevasjon.append(elev)
                        R_ij_korr.append(R_ij[index] - trop - ion)
                    else:
                        removed_indexes.append(index)

                # fjerne tilsvarende elementer fra R_ij og T_j
                for i in sorted(removed_indexes, reverse=True):
                    del R_ij[i]
                    del T_j[i]
                    del sv_list[i]

                deltaX = [v[0] - x_st0 for v in sat_koord]
                deltaY = [v[1] - y_st0 for v in sat_koord]
                deltaZ = [v[2] - z_st0 for v in sat_koord]

                A = design_matrise(deltaX, deltaY, deltaZ)

                if mixed_gnss:
                    extra_column = [[1] if sv_i[0] == 'E' else [0] for sv_i in sv_list]
                    A = np.hstack((A,extra_column))

                ant_obs = A.shape[0]
                ant_ukjente = A.shape[1]
                if ant_obs <= ant_ukjente:
                    break

                avstander = [math.sqrt(deltaX[i] ** 2 + deltaY[i] ** 2 + deltaZ[i] ** 2) for i in range(len(deltaX))]
                if mixed_gnss:
                    L = [R_ij_korr[i] + Trel_m[i] - avstander[i] - c * dt0 - c * dt_system * A[i, 4] for i in range(len(R_ij_korr))]
                else:
                    L = [R_ij_korr[i] + Trel_m[i] - avstander[i] - c * dt0 for i in range(len(R_ij_korr))]

                N = np.matmul(np.transpose(A), A)
                h = np.matmul(np.transpose(A), L)
                dX = np.matmul(np.linalg.inv(N), h)

                x_st0 += dX[0]
                y_st0 += dX[1]
                z_st0 += dX[2]
                dt0 += dX[3] / c
                if mixed_gnss:
                    dt_system += dX[4] / c

                test = [math.fabs(t) for t in dX[0:3]]
                forbedring = max(test)

            if ant_obs > ant_ukjente:
                # Beregner PDOP osv kun hvis antall satellitter er stort nok
                # Residual
                v = np.matmul(A, dX) - L
                fkv = np.matmul(np.transpose(v), v)

                s_0 = math.sqrt(fkv / (ant_obs - ant_ukjente))

                Qxx = np.linalg.inv(N)
                C_ECEF = s_0 ** 2 * Qxx

                s_x = s_0 * math.sqrt(Qxx[0, 0])
                s_y = s_0 * math.sqrt(Qxx[1, 1])
                s_z = s_0 * math.sqrt(Qxx[2, 2])
                s_dT = s_0 * math.sqrt(Qxx[3, 3])
                if mixed_gnss:
                    s_dT_system = s_0 * math.sqrt(Qxx[4, 4])

                PDOP = math.sqrt(Qxx[0, 0] + Qxx[1, 1] + Qxx[2, 2])
                TDOP = math.sqrt(Qxx[3, 3])
                GDOP = math.sqrt(PDOP ** 2 + TDOP ** 2)

                lon, lat, alt = Transformer.from_proj(ecef, lla).transform(x_st0, y_st0, z_st0, radians=True)

                east, north, height = Transformer.from_proj(lla, utm).transform(lon, lat, alt, radians=True)

                M = np.array([
                    [-math.sin(lon), math.cos(lon), 0],
                    [-math.sin(lat) * math.cos(lon), -math.sin(lat) * math.sin(lon), math.cos(lat)],
                    [math.cos(lat) * math.cos(lon), math.cos(lat) * math.sin(lon), math.sin(lat)]
                ])

                C_ENU = np.matmul(np.matmul(M, C_ECEF[:3, :3]), np.transpose(M))

                s_E = math.sqrt(C_ENU[0, 0])
                s_N = math.sqrt(C_ENU[1, 1])
                s_h = math.sqrt(C_ENU[2, 2])

                #if current_time.second == 0 and current_time.minute % 5 == 0:
                if current_time.minute == 0 and current_time.second == 0:
                    data_to_resultater.append([current_time.isoformat(), current_gps_time.seconds - dt0, x_st0, y_st0, z_st0, s_0, PDOP, s_x, s_y, s_z, north, east, height, s_N, s_E, s_h])

                # skrive ut resultater i kladdefil

                mellomresultater.write('Aktuelt tidspunkt: {:s} \n'.format(current_time.isoformat()))
                mellomresultater.write('Mottaker koordinater (stasjon {:s}):\n'.format(obs_data.header.marker_name))
                mellomresultater.write('{:s} {:.3f} {:s} \n'.format('x_mot =', x_st0, 'm'))
                mellomresultater.write('{:s} {:.3f} {:s} \n'.format('y_mot =', y_st0, 'm'))
                mellomresultater.write('{:s} {:.3f} {:s} \n'.format('z_mot =', z_st0, 'm'))

                mellomresultater.write('\n')
                mellomresultater.write('Mottakerens klokkefeil:\n')
                mellomresultater.write('{:s} {:.6f} {:s} \n'.format('dt0 = ', dt0, 'sec'))
                mellomresultater.write('{:s} {:.6f} {:s} \n'.format('dt_system = ', dt_system, 'sec'))

                mellomresultater.write('\n')
                mellomresultater.write('Utjevningskorreksjoner:\n')
                with np.printoptions(precision=3, suppress=True):
                    mellomresultater.write('{0}\n'.format(v))

                mellomresultater.write('\n')
                mellomresultater.write('Standardavvik til vektsenheten:\n')
                mellomresultater.write('{:s} {:.3f}\n'.format('s0 = ', s_0))

                mellomresultater.write('\n')
                mellomresultater.write('Standardavvik til de ukjente parameterene:\n')
                mellomresultater.write('{:s} {:.3f}\n'.format('sx = ', s_x))
                mellomresultater.write('{:s} {:.3f}\n'.format('sy = ', s_y))
                mellomresultater.write('{:s} {:.3f}\n'.format('sz = ', s_z))
                mellomresultater.write('{:s} {:.3f}\n'.format('sdT = ', s_dT))
                if mixed_gnss:
                    mellomresultater.write('{:s} {:.3f}\n'.format('s_dT_system = ', s_dT_system))

                mellomresultater.write('\n')
                mellomresultater.write('{:s} {:.3f}\n'.format('3D-nøyaktighet: PDOP = ', PDOP))
                mellomresultater.write('{:s} {:.3f}\n'.format('Tidsnøyaktighet: TDOP = ', TDOP))
                mellomresultater.write('{:s} {:.3f}\n'.format('Geometrisk nøyaktighet: GDOP = ', GDOP))

                mellomresultater.write('\n')

                for sv, index in zip(sv_list, range(len(T_j))):
                    mellomresultater.write('-----------------\n')
                    mellomresultater.write('Satellitt nummer: {:s}\n'.format(sv))

                    # koordinater
                    mellomresultater.write('{:s} {:.3f} {:s}\n'.format('Xk = ', sat_koord[index][0],'m'))
                    mellomresultater.write('{:s} {:.3f} {:s}\n'.format('Yk = ', sat_koord[index][1],'m'))
                    mellomresultater.write('{:s} {:.3f} {:s}\n'.format('Zk = ', sat_koord[index][2],'m'))
                    mellomresultater.write('\n')

                    # asimut- og elevasjonsvinkler
                    mellomresultater.write('{:s} {:.3f} {:s}\n'.format('Asimutvinkel = ', math.degrees(asimut[index]), 'deg'))
                    mellomresultater.write('{:s} {:.3f} {:s}\n'.format('Elevasjonsvinkel = ', math.degrees(elevasjon[index]), 'deg'))
                    mellomresultater.write('\n')

            current_time += interval

        resultater.write(
            tabulate(data_to_resultater,
                     headers=['epoke', 'tow-dt0', 'x_ecef', 'y_ecef', 'z_ecef', 's0', 'PDOP','s_x', 's_y', 's_z', 'N_UTM', 'E_UTM', 'h_ell', 's_N', 's_E', 's_h'],
                     floatfmt=("", "15.10f") + ("15.4f",) * 14)
        )

    return x_st0, y_st0, z_st0


if __name__ == '__main__':
    try:
        os.remove("./mellomresultater.txt")
        os.remove("./resultater.txt")
    except OSError:
        pass
    print("Mottakers koordinater: (x, y, z)")
    korr = True
    point = "ssir"
    print(
        spp(
            obs_file="../testdata!!!/%s2720.22o" % point,
            nav_file="../testdata!!!/%s2720.22p" % point,
            gnss={'G': 'C1C', 'E': 'C1X'},
            #gnss = {'E': 'C1X'},
            #gnss={'G': 'C1C'},
            start_tidspunkt=datetime(2022, 9, 29, 11, 00, 0, 0),
            slutt_tidspunkt=datetime(2022, 9, 30, 5, 00, 0, 0),
            fil_for_mellomresultater="./mellomresultater.txt",
            fil_for_hoved_resultater="./resultater/gps_galileo/%s_gps_galileo_18t_int_1t_atm.txt" % point,
            trop_korreksjoner=korr,
            ion_korreksjoner=korr
        )
    )

