import math
from datetime import datetime

import numpy as np
from gps_time import GPSTime

from geo import GNSS, c


def beregne_sv_koordinater(
        block, # block med nav data for en gitt GNSS
        tidspunkt: GPSTime,
        mot_koord: (float, float, float),
        gnss: GNSS
) -> (float, float, float, float, float):
    """
    Funksjonen for å beregne satellittskoordinater for en gitt tidspunkt.
    Metoden sjekker også om jordrotasjonskorreksjoner skal brukes i beregningene.

    :param block: block med navigasjonsdata for en git GNSS
    :param tidspunkt: GPSTime. tidspunkt i gps tid
    :param mot_koord: (float, float, float). koordinatene til mottaker.
    :param gnss: GNSS object. Se geo.py
    :return: xk, yk, zk, tk, Trel_m
    """
    block_time = datetime.strptime(block.timestamp, '%Y-%m-%dT%H:%M:%S')
    block_time_gps = GPSTime.from_datetime(block_time)

    tk = tidspunkt - block_time_gps  # time elapsed since the reference epoch

    # Beregning av satellittens eksentriske anomali

    A = block.sqrt_A ** 2  # semi-major axis
    n0 = math.sqrt(gnss.GM / A ** 3)  # computed mean motion
    n = n0 + block.Delta_n  # corrected mean motion
    Mk = block.M0 + (n * tk)  # mean anomaly

    E0 = Mk
    E_mellom = Mk + block.e * math.sin(E0)

    while math.fabs(E_mellom - E0) > 1e-15:
        E0 = E_mellom
        Ek = Mk + block.e * math.sin(E_mellom)  # eccentric anomaly
        E_mellom = Ek

    # Beregning av satellittens sanne anomali

    cos_vk = (math.cos(Ek) - block.e) / (1 - block.e * math.cos(Ek))
    sin_vk = math.sqrt(1 - block.e ** 2) * math.sin(Ek) / (1 - block.e * math.cos(Ek))
    vk = math.atan2(sin_vk, cos_vk)  # true anomaly

    Trel_sek = -(2 / c ** 2) * math.sqrt(A * gnss.GM) * block.e * math.sin(Ek)
    Trel_m = Trel_sek * c

    # Beregning av korrigerte verdier for latitude, longitude, inklinasjon og radius

    Fi_k = vk + block.omega  # argument of latitude

    korr_uk = block.Cuc * math.cos(2 * Fi_k) + block.Cus * math.sin(2 * Fi_k)  # latitude correction
    korr_rk = block.Crc * math.cos(2 * Fi_k) + block.Crs * math.sin(2 * Fi_k)  # radius correction
    korr_ik = block.Cic * math.cos(2 * Fi_k) + block.Cis * math.sin(2 * Fi_k)  # inclination correction

    uk = Fi_k + korr_uk  # corrected latitude
    rk = A * (1 - block.e * math.cos(Ek)) + korr_rk  # corrected radius
    ik = block.i0 + block.IDOT * tk + korr_ik  # corrected inclination

    xk_dot = rk * math.cos(uk)  # position in the orbital plane
    yk_dot = rk * math.sin(uk)  # position in the orbital plane

    # jordrotasjonskorreksjoner
    if math.fabs(mot_koord[0]) > 1 and math.fabs(mot_koord[1]) > 1 and math.fabs(mot_koord[2]) > 1:
        # utføres
        TRANS_0 = 0.075
        TRANS = 0
        j = 0
        while math.fabs(TRANS_0 - TRANS) > 1e-10:
            j += 1
            if j > 20:
                raise ValueError("Gangtids-rotasjonen konvergerer ikke")
            TRANS = TRANS_0
            Omega_k = block.OMEGA0 + (block.OMEGA_DOT - gnss.omega_e) * tk - gnss.omega_e * (
                        block.Toe + TRANS)  # corrected longitude

            # Beregning av satellittens koordinater i jordfast system ved tidspunkt target_time
            xk = xk_dot * math.cos(Omega_k) - yk_dot * math.sin(Omega_k) * math.cos(ik)
            yk = xk_dot * math.sin(Omega_k) + yk_dot * math.cos(Omega_k) * math.cos(ik)
            zk = yk_dot * math.sin(ik)

            # Avstander mottaker-satellitt
            dX = xk - mot_koord[0]
            dY = yk - mot_koord[1]
            dZ = zk - mot_koord[2]
            dS = math.sqrt(dX ** 2 + dY ** 2 + dZ ** 2)

            TRANS_0 = dS / c
        # end while
    else:
        # ikke utføres
        Omega_k = block.OMEGA0 + (block.OMEGA_DOT - gnss.omega_e) * tk - gnss.omega_e * block.Toe  # corrected longitude

        # Beregning av satellittens koordinater i jordfast system ved tidspunkt target_time

        xk = xk_dot * math.cos(Omega_k) - yk_dot * math.sin(Omega_k) * math.cos(ik)
        yk = xk_dot * math.sin(Omega_k) + yk_dot * math.cos(Omega_k) * math.cos(ik)
        zk = yk_dot * math.sin(ik)

    return xk, yk, zk, tk, Trel_m


def asimut_og_elevasjon(
        mottaker: (float, float, float),
        satellitt: (float, float, float),
        gnss: GNSS
) -> (float, float):
    """
    Metoden for beregning av asimut og elevasjon for en gitt SV.

    :param mottaker: (float, float, float). Koordinatene til mottaker.
    :param satellitt:  (float, float, float). Koordinatene til satellitt.
    :param gnss: korrekt GNSS objekt. Se geo.py
    :return: asimut, elevasjon
    """

    # koordinater til stasjon T836 fra ECEF til geodetisk system
    x_st = mottaker[0]
    y_st = mottaker[1]
    z_st = mottaker[2]

    p = math.sqrt(x_st ** 2 + y_st ** 2)
    lon_st = math.atan(y_st / x_st)
    k = math.atan(z_st / ((1 - gnss.geod.f) * p))

    e2 = 2 * gnss.geod.f - gnss.geod.f ** 2  # eksentrisitet ^ 2
    b = gnss.geod.a * math.sqrt(1 - e2)  # lille halvaksen

    lat_st = math.atan((z_st + (e2 / (1 - gnss.geod.f)) * gnss.geod.a * (math.sin(k)) ** 3) / (p - e2 * gnss.geod.a * (math.cos(k)) ** 3))

    N = gnss.geod.a / math.sqrt(1 - e2 * math.sin(lat_st) ** 2)

    h_st = p / math.cos(lat_st) - N

    # transformasjon til lokal ENU system med origo i T836

    m11 = - math.sin(lon_st)
    m12 = math.cos(lon_st)
    m13 = 0

    m21 = - math.sin(lat_st) * math.cos(lon_st)
    m22 = - math.sin(lat_st) * math.sin(lon_st)
    m23 = math.cos(lat_st)

    m31 = math.cos(lat_st) * math.cos(lon_st)
    m32 = math.cos(lat_st) * math.sin(lon_st)
    m33 = math.sin(lat_st)

    M = np.array([[m11, m12, m13], [m21, m22, m23], [m31, m32, m33]])

    # dkoord
    dkoord = np.array([satellitt[0] - x_st, satellitt[1] - y_st, satellitt[2] - z_st])

    # enu koordinater
    sat_koord_ENU = np.matmul(M, dkoord)

    # asimut- og elevasjonsvinkler

    ne = math.sqrt(sat_koord_ENU[1] ** 2 + sat_koord_ENU[0] ** 2)
    elevasjon = math.atan2(sat_koord_ENU[2], ne)  # rad
    asimut = math.atan2(sat_koord_ENU[0], sat_koord_ENU[1])  # rad

    if elevasjon < 0:
        elevasjon = elevasjon + 2 * math.pi

    if asimut < 0:
        asimut = asimut + 2 * math.pi

    return asimut, elevasjon


def design_matrise(
        dx, dy, dz
):
    """
    Metoden for opprettelse av design matrise A.
    :param dx: vektor dX
    :param dy: vektor dY
    :param dz: vektor dZ
    :return: design matrise A
    """
    ant_sat = len(dx)

    A = np.zeros((ant_sat, 4))

    for i in range(ant_sat):
        avst = math.sqrt(dx[i] ** 2 + dy[i] ** 2 + dz[i] ** 2)
        A[i, 0] = -dx[i] / avst
        A[i, 1] = -dy[i] / avst
        A[i, 2] = -dz[i] / avst
        A[i, 3] = 1

    return A


def beregne_R_ji(
        nav_block, obs_block, sv, current_gps_time, current_time
):
    """
    Metoden for beregning av verdier, som inngår i R_ij vektor.
    :param nav_block: korrekt data block fra RINEX navigasjonsfil.
    :param obs_block: korrekt data block fra RINEX observasjonsfil.
    :param sv: satellittnavn
    :param current_gps_time: GPSTime. tidspunkt i GPS format
    :param current_time: datetime. tidspunkt i datetime format.
    :return: float. Verdi for r_ij
    """
    a0 = nav_block.clock_bias
    a1 = nav_block.clock_drift
    a2 = nav_block.clock_drift_rate
    T_GD = nav_block.TGD
    toc = nav_block.Toe
    dT_j = a0 + a1 * (current_gps_time.seconds - toc) + a2 * (current_gps_time.seconds - toc) ** 2 - T_GD

    r_ij = obs_block[sv][current_time.isoformat()][0]["value"]
    return r_ij + c * dT_j